;; Generated package description from jit-spell.el  -*- no-byte-compile: t -*-
(define-package "jit-spell" "0.5" "Just-in-time spell checking" '((emacs "27.1") (compat "29.1")) :commit "e6f3525b848c48906f06bf21a9e8678b95ccd3bf" :authors '(("Augusto Stoffel" . "arstoffel@gmail.com")) :maintainer '("Augusto Stoffel" . "arstoffel@gmail.com") :keywords '("tools" "wp") :url "https://github.com/astoff/jit-spell")
